theory Submission
  imports Defs
begin

lemma bij_betw_If:
  assumes "bij_betw f A B" "bij_betw g C D" "A \<inter> C = {}" "B \<inter> D = {}"
  shows   "bij_betw (\<lambda>x. if x \<in> A then f x else g x) (A \<union> C) (B \<union> D)"
  using assms by (auto simp: bij_betw_def inj_on_def)

lemma bij_betw_id'[intro]: "bij_betw (\<lambda>x. x) A A"
  by (auto simp: bij_betw_def)

text \<open>Task 2 - 4/5 points\<close>
theorem bijective_embedding:
  fixes f :: "nat \<Rightarrow> nat"
    and S :: "nat set"
  assumes "inj_on f S" and "finite S"
  shows "\<exists>h. bij_betw h \<nat> \<nat> \<and> (\<forall>x \<in> S. h x = f x)"
proof -
  text \<open>
    Let \<open>N\<close> be the biggest number occurring in the mapping \<open>f\<close> (in either the domain
    or the range). Then we can simply look at the finite domain \<open>{0..N}\<close> and map everything
    else to itself. On \<open>{0..N}\<close>, we can find a bijection between \<open>{0..N} - S\<close> and \<open>{0..N} - f ` S\<close>
    because they obviously have the same cardinality. We thus have bijections between \<open>S\<close> and
    \<open>f ` S\<close>, between \<open>{0..N} - S\<close> and \<open>{0..N} - f ` S\<close>, and between \<open>{N<..}\<close> and \<open>{N<..}\<close>.

    Pasting these three together we obtain a bijection from \<open>\<nat>\<close> to \<open>\<nat>\<close>.
  \<close>
  define N where "N = Max (S \<union> f ` S)"
  have "card ({..N} - S) = Suc N - card S"
    using assms by (subst card_Diff_subset) (auto simp: N_def)
  also have "\<dots> = card ({..N} - f`S)"
    using assms by (subst card_Diff_subset) (auto simp: N_def card_image)
  finally have "\<exists>g. bij_betw g ({..N} - S) ({..N}- f ` S)"
    by (intro finite_same_card_bij) auto
  then guess g .. note g = this

  define h where "h = (\<lambda>x. if x \<in> S then f x else (if x \<in> {..N} - S then g x else x))"
  have "bij_betw h (S \<union> (({..N} - S) \<union> {N<..})) (f ` S \<union> (({..N} - f ` S) \<union> {N<..}))"
    unfolding h_def using assms by (intro bij_betw_If g inj_on_imp_bij_betw) (auto simp: N_def leD)
  also have "S \<union> (({..N} - S) \<union> {N<..}) = \<nat>"
    by (auto simp: Nats_def)
  also have "f ` S \<union> (({..N} - f ` S) \<union> {N<..}) = \<nat>"
    by (auto simp: Nats_def)
  finally have "bij_betw h \<nat> \<nat>" .
  moreover have "\<forall>x\<in>S. h x = f x" by (auto simp: h_def)
  ultimately show ?thesis by blast
qed

text \<open>Task 1 - 1/5 points\<close>
theorem injective_embedding:
  fixes f :: "nat \<Rightarrow> nat"
    and S :: "nat set"
  assumes "inj_on f S" and "finite S"
  shows "\<exists>g. inj_on g \<nat> \<and> (\<forall>x \<in> S. g x = f x)"
  using bijective_embedding[OF assms] by (auto simp: bij_betw_def)

end
